import argparse
import lazyconf

from lib.schema import *
from lib.prompt import *
from lib.select import *
from lib.merge import *

# A function to be called from the setup entrypoint.
def parse():
    parser = argparse.ArgumentParser(prog='lazyconf')
    sp = parser.add_subparsers(dest='command')

    sp_config = sp.add_parser('config', help='Run the configurator.')
    
    sp_get = sp.add_parser('get', help='Gets the value for a specific key in the configuration.')
    sp_set = sp.add_parser('set', help='Sets the value for a specific key in the configuration.')
    sp_parse = sp.add_parser('parse', help='Parses and saves previously set keys')
    
    sp_get.add_argument('-k', '--key',
                        help='The key to retrieve.', dest='key', required=True)
    sp_set.add_argument('-k', '--key',
                        help='The key to set for.', dest='key', required=True)
    sp_set.add_argument('-v', '--value',
                        help='The value to set.', dest='value', required=True)
    
    return parser.parse_args()

def get(l, key):
    try:
        l.load()
    except:
        exit()
    
    r = l.get(key)
    if r:
        print r
    exit()

def set(l, key, value):
    try:
        l.load()
    except Exception as e:
        raise  e

    try:
        l.set(key, value)
    except Exception as e:
        raise e

    exit()

def console():
    command_dict = parse().__dict__
    command = command_dict['command']

    l = lazyconf.Lazyconf()

    if command == 'get':
        key = command_dict['key']
        get(l, key)
    elif command == 'config':
        l.configure()
    elif command == 'set':
        key = command_dict['key']
        value = command_dict['value']
        set(l, key, value)
       
# Call the console function if we call this file directly.
if __name__ == '__main__':
    console()




""" Merge used to merge variables in the schema dictionary into the data dictionary, without losing defaults.
    It never returns the dictionaries, as the original dictionaries are passed by reference automatically.
    The schema dictionary is never be modified by this class. """

class Merge():

    # Initialisation creates sets of the top level keys in both dictionaries, and the intersection between the two.
    def __init__(self, schema, data):
        self.schema = schema
        self.data = data

        # Create the sets.
        self.set_schema = set(self.schema.keys())
        self.set_data = set(self.data.keys())
        self.intersect = self.set_schema.intersection(self.set_data)
        

    # Returns all top level keys which have been added to the schema.
    def added(self):
        return self.set_schema - self.intersect


    # Returns all top level keys which have been removed from the schema.
    def removed(self):
        return self.set_data - self.intersect

    # This method recursively merges all changes in the schema into the data file.
    def merge(self, schema = None, data = None, prefix = None):
        
        # If there's no prefix yet, don't add a '.'
        p = ''
        if prefix:
            p = prefix + '.'
        
        # Create the dictionary to be returned.
        mods = {
            'added'   : [], # []string
            'removed' : [], # []string
            'modified': []  # [](string,(type,type))
        }

        if schema is None:
            schema = self.schema

        if data is None:
            data = self.data

        # Add all new keys in schema to data.
        added = self.added()
        if added:
            for a in added:
                mods['added'].append(p + a)
                data[a] = schema[a]

        # Remove all keys in data that are no longer in schema.
        removed = self.removed()
        if removed:
            for r in removed:
                mods['removed'].append(p + r)
                del(data[r])

        # Recursively call this method until all common variables are resolved.
        intersect = self.intersect
        if intersect:
            for i in intersect:
                s = type(schema[i])
                d = type(data[i])

                # If there is a dictionary that is shared at this key by both the schema and the data, call this method again.
                if s is dict and d is dict:
                    
                    # Create a new instance of this class with the child dicts as schema and data.
                    diff = Merge(schema[i], data[i])
                    
                    # Merge the child dicts.
                    add_mods = diff.merge(prefix = p + i)

                    # Update our return dict.
                    mods['added'] += add_mods['added']
                    mods['removed'] += add_mods['removed']
                    mods['modified'] += add_mods['modified']
                    
                    # Remove the created Merge.
                    del(diff)

                # Otherwise if the variable type has changed, copy the variable from the schema to the data.
                elif s != d:
                    
                    # Here, we process lists into empty strings.
                    if s is list:
                            data[i] = ""

                    # If we're not dealing with a list, process normally.
                    else: 
                        # Add this to the 'modified' list.
                        m = (p + i, (str(d),str(s)))
                        mods['modified'].append(m)

                        # Copy the value from the schema to the data.
                        data[i] = schema[i]

        # Return the dictionary of modifications.
        return mods


import re
from colors import Colors

""" Prompt contains all the methods for parsing and returning input. """

class Prompt():

    # Initialisation
    def __init__(self):
        pass

    ### Formatting ###

    # Prints a header around a string.
    def header(self, msg):
        print(Colors.header + msg + Colors.end)


    # Prints a success message.
    def success(self, msg):
        print(Colors.success + msg + Colors.end)


    # Prints an error message.
    def error(self, msg):
        print(Colors.error + msg + Colors.end)


    # Prints a notice message.
    def notice(self, msg):
        print(Colors.blue + msg + Colors.end)


    def validate_prompt(self, value, validate):
        if validate:
            if callable(validate):
                try:
                    value = validate(value)
                except Exception, e:

                    # If the regex fails, loop.
                    value = None
                    self.error("Validation failed for the following reason:")
                    self.error(e.message + '\n')

            # Attempt to validate it if it's not callable.
            else:
                if not validate.startswith('^'):
                    validate = r'^' + validate

                if not validate.endswith('$'):
                    validate += r'$'

                result = re.findall(validate, value)
                if not result:
                    self.error("Regular expression validation failed: '%s' does not match '%s'\n" % (value, validate))

                    # If the regex fails, loop.
                    value = None

        return value

    # The prompt string method.
    # This method heavily models how it works in Fabric.
    def prompt(self, label, value=None, default='', validate=None):
        
        p = label + ' ' + '[%s]' % str(default).strip() + ' '
        prompt_string = None

        if not value:
            value = None
            while value is None:
                prompt_string = raw_input(p) or default
                value = self.validate_prompt(prompt_string, validate)
        else:
            value = self.validate_prompt(value, validate)

        return value

    ### Prompts ###

    # Returns the value from a validated int prompt.
    def int(self, label, value=None, default = 0):
        ret = self.prompt(label + ' (int):', value=value, default = str(default), validate = r'^[0-9]+$')
        return int(ret)


    # Returns the value from a validated bool prompt.
    def bool(self, label, value=None, default = False):
        ret = self.prompt(label + ' (y/n):', value=value, default = self.fmt_bool(default), validate = r'^(y|n)$')
        return self.defmt_bool(ret)


    # Returns the value from a validated select prompt.
    def select(self, label, select, value=None, default = ""):
        dflt = select.get_key(default)
        if not dflt:
            dflt = select.first_choice()

        return select.get_value(self.prompt(label + ' ' + select.choices() + ':', value = value, default = dflt, validate = select.reg_choices()))

    ### Formatting ###

    # Takes a python bool and returns it as y|n.
    def fmt_bool(self, p):
        if p is True:
            return 'y'
        if p is False:
            return 'n'
        return ''


    # Takes y|n and returns a python bool.
    def defmt_bool(self, p): 
        if p == 'y':
            return True
        if p == 'n':
            return False
        return None